/**
 * MIT License
 *
 * Copyright (c) [year] [fullname]
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 * @license MIT
 */

/**
 * Provides utility functions to test, normalize and parse strings formatted as
 * European Case Law Identifier (ECLI) version 1.0. By default the parsing
 * function will operate in strict mode and require strict adherence to the
 * spec. Optionally, the parsing function can operate in lenient mode, meaning
 * that all fields except the id field may be omitted, in which case they are
 * repliced by default values provided to the parser. The test and normalize
 * function always operate in strict mode.
 * @module
 */

/**
 * Regexp literal for ECLI 1.0 strings. The regex follows the format
 * proposed in the Annex of EU Council Conclusion 2011/C 127/01 of 29 April
 * 2011. The regex matches the formatting but does not validate the individual
 * fields. For example, the second field matches if it contains two letters, but
 * the regex does not validate its contents as a valid EU or ISO 3166 alpha-2
 * country code. As per the spec, the regex is case-insensitive.
 * @constant
 * @type {RegExp}
 * @private
 */
const ECLI_REGEX = /^(ECLI):([A-Za-z]{2}):([A-Za-z]{1}[A-Za-z0-9]{0,6}):([0-9]{4}):([A-Za-z0-9.]{1,25})$/i

/**
 * The number of fields in an ECLI 1.0 (5).
 * @constant
 * @type {Number}
 * @private
 */
const ECLI_FIELD_COUNT = 5;

/**
 * The maximum number of characters in a valid ECLI 1.0 string (46).
 * @constant
 * @type {Number}
 * @private
 */
const ECLI_MAX_CHARS = 46;

/**
 * The character used to separate fields in an ECLI 1.0 string (':').
 * @constant
 * @type {String}
 * @private
 */
const ECLI_FIELD_SEP = ':';

/**
 * A simple Regexp to check for four-digit integers, which for the sake of
 * simplicity we will accept for the year field in the ECLI strings.
 * @constant
 * @type {Regexp}
 * @private
 */
const YEAR_REGEX = /^[0-9]{4}$/

/**
 * Return true if the input represents a valid formatted ECLI 1.0. The test
 * is always performed in strict mode. To test in lenient mode, call the parse()
 * function instead and test the returned object for nullity.
 * @function test
 * @param {String} str - A string to test for compliance with the ECLI 1.0
 *  specification.
 * @returns {Boolean} True if the input complies with the ECLI 1.0 spec, false
 *  otherwise.
 */
module.exports.test = str => ECLI_REGEX.test(str)

/**
 * Return the input string in uppercase if it is a valid formatted ECLI 1.0,
 * otherwise return null. Operates only in strict mode. To normalize in lenient
 * mode, call the parse() function. If the returned object is not null, it will
 * contain a string property called text that contains the full, normalized
 * ECLI.
 * @function normalize
 * @param {String} str - A string to test and uppercase.
 * @returns {String} The input string in uppercase if it is a valid ECLI 1.0,
 *  otherwise null.
 */
module.exports.normalize = str => {
  return (ECLI_REGEX.test(str) ? str.toUpperCase() : null)
}

/**
 * Return a parsing function with the specified configuration options. If opts
 * is missing, a strict parser is returned that parses only ECLI's that fully
 * match the ECLI 1.0 specification. Otherwise, the returned function will parse
 * leniently and apply the default values provided in the opts argument if the
 * corresponding fields are missing in the input string. The parse function
 * accepts a single string argument and returns null if the input string could
 * not be parsed as an ECLI 1.0. If the parse succeeds, the function returns an
 * object with the ECLI's individual fields as properties.
 * @function create
 * @param {Object} [opts={}] - Config options for the parser to be created.
 * @param {Boolean} [opts.lenient] - If true, the parser will accept shorthand
 *  input strings, without the ECLi prefix and country, court and/or year
 *  fields. Default values for these fields should be provided in the opts
 *  argument.
 * @param {String} [opts.country] - The default value for the ECLI country
 *  field. Must be set if the court property is also set.
 * @param {String} [opts.court] - The default value for the ECLI court field. If
 *  this property is set, the country property must also be set, otherwise
 *  a RangeError is thrown.
 * @param {Number} [opts.year] - The default value for the ECLI year field.
 *  If not set, the lenient parser will use the current year.
 * @throws {RangeError} - Throws a RangeError if the court property is set but
 *  the country property is not.
 * @throws {TypeError} - Throws a TypeError if the year property is not a four
 *  digit integer.
 * @returns {Function} A function that accepts a single string parameter and
 *  returns an object with properties corresponding to the individual ECLI
 *  fields of the input string (country, court, year, id). The object also
 *  contains a string property called text which contains the full ECLI in
 *  uppercase and a property called version with is set to the string '1.0'.
 *  If the input string cannot be parsed as a valid ECLI 1.0, the function
 *  returns null.
 */
module.exports.create = function(opts = {}) {

  // Only leading fields may be omitted, no gaps in the middle of an ECLI. For
  // example, '2017:148' is valid, as the ECLI, the country and the court can
  // be supplied as default values, but 'NL::2017:148' is invalid, because it
  // contains a gap. So if a default value for the court is provided, a default
  // value for the country must also be provided. Technically, this should apply
  // to the year field as well (i.e. country and court are required if a default
  // is provided for the year field), but if omitted the year defaults to the
  // current year.
  if(opts.lenient && (opts.court && !opts.country)) {
    const msg = 'Ecli-parser option country must be set because court is set.'
    throw new RangeError(msg)
  }

  if(!opts.year) {
    opts.year = (new Date()).getFullYear()
  } else if (!YEAR_REGEX.test(opts.year)) {
    throw new TypeError('Bad value for ecli-parser option year: ' + opts.year)
  }

  return function(str) {

    // A trailing ':' screws up the call to String.split() below. Since a
    // trailing ':' means the input is invalid anyway, we simply check for it
    // explicitly. While we're at it, reject empty strings, nulls and oversized
    // strings immediately as well (with the added benefit of avoiding a
    // ReferenceError if the input is null or undefined).
    if(!str || (str.length > ECLI_MAX_CHARS) || str.endsWith(ECLI_FIELD_SEP)) {
      return null
    }

    // First attempt a (strict) regexp match and return the result if it
    // succeeds or if lenient mode is not enabled.
    const ecli = matchRegex(str)
    if (ecli || !opts.lenient) {
      return ecli
    }

    const fields = str.split(ECLI_FIELD_SEP)

    // A leading colon is acceptable for a lenient parse. If present, the first
    // array item will be an empty string that can simply be discarded.
    if(!fields[0]) {
      fields.shift()
    }

    // If the number of items in the fields array is out of bounds or if the
    // array contains empty items at this point, the input is definitely
    // invalid.
    if (fields.length == 0 ||
        fields.length > ECLI_FIELD_COUNT ||
        !fields.every(x => x)) { // checks for falsy items in the array
      return null
    }

    // Build a new ECLI string by prepending the required number of default
    // values to the fields array. The result could still be an invalid ECLI
    // string, but the subsequent regexp match will catch that.
    const defs = ['ECLI', opts.country, opts.court, opts.year]
    const ecli2 = defs.slice(0, (ECLI_FIELD_COUNT - fields.length))
                      .concat(fields)
                      .join(ECLI_FIELD_SEP)

    // At this point, the strict regex match should succeed. If not, the input
    // was invalid to begin with.
    return matchRegex( ecli2 )
  }
}

/**
 * Parse the input string as a full ECLI 1.0 by matching against ECLI_REGEX
 * and returning an object with the ecli's individual fields as properties.
 * Returns null if the input string does not match the regex.
 * @function matchRegex
 * @param {String} input - An ECLI 1.0 formatted string. Should not be null or
 *  undefined.
 * @returns {Object} An object with the individual fields of the ECLI as its
 *  properties, i.e. country, court, year and id. The year property is an
 *  integer, the other properties are uppercase strings. The object also has a
 *  property called text, which contains the entire ECLI as an uppercase
 *  string. Finally, the object has a property called version, which contains
 *  the value '1.0'. If the input string is not a valid formatted ECLI 1.0
 *  the function returns null.
 * @private
 */
function matchRegex(input) {

  const matches = ECLI_REGEX.exec( input.toUpperCase() )

  if(!matches) {
    return null
  }

  return {
    country: matches[2],
    court: matches[3],
    year: parseInt(matches[4], 10),
    id: matches[5],
    text: matches[0],
    version: '1.0'
  }
}
